#ifndef MP_H
#define MP_H

#include <fstream.h>
#include <stdio.h>
#include "mpfun.h"
#include "mpdpe.h"

class mp_real;
class mp_int;

/**  
* This class represents mp objects.
*
* @author Herman Harjono
* @version Sep 21, 1998.
*/ 
class mp: public mpfun
{
 
 public:
   /****************** PUBLIC METHODS ********************/
/** 
  *  Constructor. Intiliaze number to 0.
  *  @param mpnw initial value of mpnw. If initialized to 
  *  a value < 1, mantissa array won't be allocated.
  *  @see #mantissa
  */
  mp(const mpsize& maxNW=mp2);
  
  /**
  *  Creates a mp (initialized to zero). 
  *  @param precision The precision level of the object, in digits.
  *  @see #mpipl
  */
  mp(const mpprecision& precision);
  
  /**
  *  Creates a mp from int. 
  *  @param precision The precision level of the object, in digits. 
  *  @see mpipl
  */
  mp(int ia, const mpprecision& precision=mpipl);
  
  /**
  *  Creates a mp from float / double. 
  *  @param precision The precision level of the object, in digits. 
  *  @see mp#mpipl
  */
  mp(double, const mpprecision& precision=mpipl);
  
  /**
  *  Creates a mp from char*. 
  *  @param precision The precision level of the object, in digits. If equals "", 
  *  it is assigned the value of mpipl.
  *  @see mp#mpipl
  */
  mp(char *, const mpprecision& precision=mpipl);
  
  /**
  *  Copy constructor. Calls mpeq.
  */
  mp(const mp&);
  
  /**
  *  Destructor. Deleting the mantissa array.
  *  @see #mantissa
  */
  virtual ~mp() {delete [] mantissa;} 


  /**
  *  Equality.
  */
  friend int operator==(const mp&, const mp&);
  
  /**
  *  Inequality.
  */
  friend int operator!=(const mp&, const mp&);
  
  /**
  *  Less than or equal to.
  */
  friend int operator<=(const mp&, const mp&);
  
  /**
  *  Greater than or equal to.
  */
  friend int operator>=(const mp&, const mp&);
  
  /**
  *  Greater than.
  */
  friend int operator<(const mp&, const mp&);
  
  /**
  *  Less than.
  */
  friend int operator>(const mp&, const mp&);
  
  /**
  *  Conversion operator to double (primitive).
  */
  friend double dble(const mp&);
  
  /**
  *  Conversion operator to int (primitive).
  */
  friend int integer(const mp&);
  
  /**
  *  Conversion operator to float (primitive).
  */
  friend float real(const mp&);
  
  /**
  *  Extraction from an input stream.
  */
  friend istream& operator>>(istream&, mp&);
  
  /**
  *  Insertion to an output stream.
  */
  friend ostream& operator<<(ostream&, const mp&);
  
  /**
  *  Sign set method.
  */
  inline void setSign(int s) {sign = s; }
  
  /**
  *  Sign get method.
  */
  inline int getSign() const {return sign;}
  
  inline mp_dpe toDPE() const
  { mp_dpe res; mpmdc(*this, res); return res;}

    /**
  *  Initializes mpl02.
  */
  static mp_real& init_mpl02();
  
  /**
  *  Initializes mpl10.
  */
  static mp_real& init_mpl10();
  
  /**
  *  Initializes mppic.
  */
  static mp_real& init_mppic();
  
  /**
  *  Initializes mpeps.
  */
  static mp_real& init_mpeps();
  

#if DEBUG == 2
/**  
*  Helper function for debugging.
  */
  static void debug(char* functionName, const mp& mpVar);
#endif
    
 protected:  
  /********************** DATA MEMBERS ***********************/

 /**
  *  The maximum number of mantissa words. 
  *  
  *  This is the size of the mantissa array.
  *  Invariant: maxnw - 2 >= nw.
  */
  int maxnw;
  
  /**
  *  The sign.  
  */
  int sign;
  
  /**
  *  The number of significant words.
  *  
  *  Invariant: nw <= maxnw - 2.
  */
  int nw;
  
  /**
  *  Exponent
  */
  int exponent;
  
  /**
  *  Mantissa array. Size of the array is maxnw.
  *  @see #maxnw
  */
  float* mantissa;   
  
  /**
  *  Subscript operator. For non-const objects.
  *  @return reference to the mantissa element.
  */
  inline float& operator[](int index){return mantissa[index];}
  
  /**
  *  Subscript operator. For const objects.
  *  @return the mantissa element.
  */  
  inline float operator[](int index) const {return mantissa[index];}
   
  /***********************  MP ARITHMETIC FUNCTIONS ******************************/
  
  /** 
  *  This routine adds mp numbers a and b to yield the mp sum c.  
  *
  *  It attempts to include all significance of a and b in the result, up to the maximum
  *  mantissa length mpnw.  Debug output starts with MPIDB = 9. This is a new
  *  simplified version. 
  *  <br>
  *  Dimension a(mpnw), b(mpnw), c(mpnw+2).
  */
  static void mpadd (const mp& a, const mp& b, mp& c, int mpnw);
  
  /**
  *  This computes the cube root of the MP number A and returns the MP result
  *  in B.  
  *
  *  For extra high levels of precision, use MPCBRX.  Debug output
  *  starts with MPIDB = 7.
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2).
  *  <br>
  *  This subroutine employs the following Newton-Raphson iteration, which
  *  converges to A ^ (-2/3):
  *  <pre>  
  *  X_{k+1} = X_k + (1 - X_k^3 * A^2) * X_k / 3
  *  </pre>
  *  where the muliplication () * X_k is performed with only half of the
  *  normal level of precision.  These iterations are performed with a
  *  maximum precision level MPNW that is dynamically changed, doubling with
  *  each iteration.  The final iteration is performed as follows (this is
  *  due to A. Karp):
  *  <pre>
  *  Cbrt(A) = (A * X_n) + [A - (A * X_n)^3] * X_n / 3 (approx.)
  *  </pre>
  *  where the multiplications A * X_n and [] * X_n are performed with only
  *  half of the final level of precision.  See the comment about the parameter
  *  NIT in MPDIVX.
  *  @see #mpcbrx
  *  @see #mpidvx
  */
  static void mpcbrt (const mp& a, mp& b, int mpnw);
  
  /**
  *  This routine compares the MP numbers A and B and returns in IC the value
  *  -1, 0, or 1 depending on whether A < B, A = B, or A > B.  
  *  
  *  It is faster than merely subtracting A and B and looking at the SIGN of the result.
  *  Debug output begins with MPIDB = 9.
  *  <br>
  *  Dimension a(mpnw+2), b(mpnw+2).
  */
  static void mpcpr (const mp&a, const mp&b, int& ic, int mpnw);
  
  
  /**
  *  This divides the MP number A by the MP number B to yield the MP quotient c.
  *
  *  For extra high levels of precision, use MPDIVX.  Debug output starts with
  *  MPIDB = 8.
  *  <br>
  *  Dimension a(mpnw), b(mpnw), c(mpnw+2).
  *  @see #mpdivx
  */
  static void mpdiv (const mp& a, const mp& b, mp& c, int mpnw);
  
  /** 
  *  This routine divides the MP number A by the DPE number (B, N) to yield
  *  the MP quotient c.  
  *
  *  Debug output starts with MPIDB = 9.
  *  <br>
  *  Dimension a(mpnw), c(mpnw+2).
  */
  static void mpdivd(const mp& a,const mp_dpe& b,  
    mp& c, int mpnw);
  
    /**
    *  This routine converts the DPE number (A, N) to MP form in B.  
    *  
    *  All bits of A are recovered in B.  However, note for example that if A = 0.1 and N
    *  is 0, then B will NOT be the multiprecision equivalent of 1/10.  Debug
    *  output starts with MPIDB = 9.
    *  <br>
    *  Dimension b(6).
  */
  static void mpdmc(const mp_dpe& a, mp& b);
  
  /**
  *  This routine sets the MP number B equal to the MP number A. 
  *
  *  Debug output starts with MPIDB = 10.
  *  <br>
  *  Dimension a(mpnw + 1), b(mpnw + 1).
  *  <br>
  *  The fact that only MPNW + 1 cells, and not MPNW + 2 cells, are copied is
  *  important in some routines that increase the precision level by one.
  */
  static void mpeq(const mp& a, mp& b, int mpnw);
  
  /**
  *  Sets B to the integer part of the MP number A and sets c equal to the
  *  fractional part of A.  
  *
  *  Note that if A = -3.3, { B = -3 and c = -0.3.
  *  Debug output starts with MPIDB = 9.
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2), c(mpnw+2).
  */
  static void mpinfr (const mp& a, mp& b, mp& c, int mpnw); 
  
  /**
  *  This converts the MP number A to the DPE number b, accurate to between
  *  14 and 17 digits, depending on system. 
  *
  *  B will be between 1 and MPBDX.
  *  Debug output starts with MPIDB = 9.
  *  dimension a(mpnw)
  */
  static void mpmdc(const mp& a, mp_dpe& b);
  
  /**
  *  This routine multiplies MP numbers A and B to yield the MP product c.
  *
  *  When one of the arguments has a much higher level of precision than the
  *  other, this routine is slightly more efficient if A has the lower level of
  *  precision.  For extra high levels of precision, use MPMULX.  Debug output
  *  starts with MPIDB = 8. 
  *  <br>
  *  Dimension a(mpnw), b(mpnw), c(mpnw+2).
  *  <br>
  *  This routine returns up to MPNW mantissa words of the product.  If the
  *  complete double-long product of A and B is desired (for example in large
  *  integer applications), { MPNW must be at least as large as the sum of
  *  the mantissa lengths of A and B.  In other words, if the precision levels
  *  of A and B are both 64 words, { MPNW must be at least 128 words to
  *  obtain the complete double-long product in c.
  *  @see #mpmulx
  */  
  static void mpmul(const mp& a, const mp&b, mp& c, int mpnw);
  
  /**
  *  This routine multiplies the MP number A by the DPE number (B, N) to yield
  *  the MP product c. 
  *
  *  Debug output starts with MPIDB = 9.  
  *  <br>
  *  Dimension a(mpnw), c(mpnw+2).
  */
  static void mpmuld (const mp& a, const mp_dpe& b, mp& c, int mpnw);
  
  /**
  *  This sets B equal to the integer nearest to the MP number A.  
  *
  *  Debug output starts with MPIDB = 8.
  *  <br>
  *  Dimension a(mpnw+2), b(mpnw+2).
  */
  static void mpnint (const mp& a, mp& b, int mpnw);
  
  
  /**
  *  This computes the N-th power of the MP number A and returns the MP result
  *  in B. 
  *
  *  When N is zero, 1 is returned.  When N is negative, the reciprocal
  *  of A ^ |N| is returned.  For extra high levels of precision, use MPNPWX.
  *  Debug output starts with MPIDB = 7.  
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2).
  *  <br>
  *  This routine employs the binary method for exponentiation.
  *  @see #mpnpwx
  */
  static void mpnpwr (const mp& a, int n, mp& b, int mpnw);
  
  /**
  *  This computes the N-th root of the MP number A and returns the MP result
  *  in B.  
  *  
  *  N must be at least one and must not exceed 2 ^ 30.  For extra high
  *  levels of precision, use MPNRTX.  Debug output starts with MPIDB = 7.  
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2).
  *  <br>
  *  This subroutine employs the following Newton-Raphson iteration, which
  *  converges to A ^ (-1/N): 
  *  <pre>
  *   X_{k+1} = X_k + (X_k / N) * (1 - A * X_k^N)
  *  </pre>
  *  The reciprocal of the final approximation to A ^ (-1/N) is the N-th root.
  *  These iterations are performed with a maximum precision level MPNW that
  *  is dynamically changed, approximately doubling with each iteration.
  *  See the comment about the parameter NIT in MPDIVX. 
  *  <br>
  *  When N is large and A is very near one, the following binomial series is
  *  employed instead of the Newton scheme: 
  *  <pre>
  *  (1 + x)^(1/N)  =  1  +  x / N  +  x^2 * (1 - N) / (2! N^2)  +  ...
  *  </pre>
  *  See the comment about the parameter NIT in MPDIVX.
  *  @see #mpnrtx
  *  @see #mpdivx
  */  
  static void mpnrt (const mp& a, int n, mp& b, int mpnw);
  
  /**
  *  This returns a pseudo-random MP number A between 0 and 1. 
  *
  *  Debug output starts with MPIDB = 9.
  *  <br>
  *  Dimension a(mpnw+2).
  */
  static void mprand (mp& a, int mpnw);
  
  /**
  *  This computes the square root of the MP number A and returns the MP result
  *  in B.
  *  
  *  For extra high levels of precision, use MPSQRX.  Debug output
  *  starts with MPIDB = 7. 
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2).
  *  <br>
  *  This subroutine employs the following Newton-Raphson iteration, which
  *  converges to 1 / Sqrt(A):  
  *  <pre>
  *   X_{k+1} = X_k + 0.5 * (1 - X_k^2 * A) * X_k
  *  </pre>   
  *  where the muliplication () * X_k is performed with only half of the
  *  normal level of precision.  These iterations are performed with a
  *  maximum precision level MPNW that is dynamically changed, doubling with
  *  each iteration.  The final iteration is performed as follows (this is
  *  due to A. Karp):
  *  <pre>
  *   Sqrt(A) = (A * X_n) + 0.5 * [A - (A * X_n)^2] * X_n  (approx.)
  *  </pre>
  *  where the multiplications A * X_n and [] * X_n are performed with only
  *  half of the final level of precision.  See the comment about the parameter
  *  NIT is MPDIVX.
  *  @see #mpsqrx
  *  @see #mpdivx
  */
  static void mpsqrt(const mp& a, mp& b, int mpnw);
  
  /**
  *  This routine subtracts MP numbers A and B to yield the MP difference c,
  *  by negating B and adding.  
  *
  *  Debug output starts with MPIDB = 9.
  *  <br>
  *  Dimension a(mpnw), b(mpnw), c(mpnw+2).
  */
  static void mpsub (const mp& a, const mp& b, mp& c, int mpnw);
  
  
  /***********************  IO & DEBUG FUNCTIONS ******************************/
#if DEBUG == 2
  /**
  *  This outputs the character string CS, the exponent of the MP number A, and
  *  the first 70 digits of A, all on one line.
  *
  *  CS must either be a literal string not exceeding 12 characters in length.
  *  <br>
  *  Dimension a(mpnw).
  */
  static void mpdeb (char* cs, const mp& a, int mpnw);
#endif
  
  /**
  *  Converts the char array A of length N into the MP number B.  
  *
  *  The string A must be in the format '10^s a x tb.c' where a, b and c are digit
  *  strings; s and t are '-', '+' or blank; x is either 'x' or '*'.  Blanks may
  *  be embedded anywhere.  The digit string a is limited to nine digits and
  *  80 total characters, including blanks.  The exponent portion (i.e. the
  *  portion up to and including x) and the period may optionally be omitted.
  *  Debug output starts with MPIDB = 7.
  *  <br>
  *  Dimension a(n), b(mpnw+2).
  */
  static void mpinpc (const char a[], int n, mp& b, int mpnw);
  
  
  /**
  *  Converts the MP number A into character form in the char  array B.
  *
  *  N (an output parameter) is the length of the output.  In other words, B is
  *  contained in B[0], ..., B[N-1].  The format is analogous to the Fortran
  *  exponential format (E format), except that the exponent is placed first.
  *  Debug output starts with MPIDB = 7.
  *  <br>
  *  Dimension a(mpnw), b(7.225 * MPNW + 30).
  *  <br>
  *  This routine is called by MPOUT, but it may be directly called by the user
  *  if desired for custom output. 
  *  @see #mpout
  *  @see #mpoutx
  */
  static void mpoutc(const mp& a, char* b, int& n, int mpnw);
  
  /**
  *  This function is a wrapper for mpoutc. 
  *  
  *  cs is the output char array.  cs must be dimensioned at least la + 25. 
  *  A comma is placed at the end of the last line to denote the end of
  *  the mp number. Here is an example of the output:
  *  <pre>
  *   10 ^        -4 x  3.14159265358979323846264338327950288419716939937510,
  *  </pre>
  *  @see #mpoutc
  */
  static void mpouts (const mp& a, int la, char* cs, int& l, int mpnw);
  
  
  /***************** ALGEBRAIC & TRANSCEDENTAL ROUTINES ***********************/
  
  /**
  *  This computes the MP angle A subtended by the MP pair (X, Y) considered as
  *  a point in the x-y plane.  
  *
  *  This is more useful than an arctan or arcsin
  *  routine, since it places the result correctly in the full circle, i.e.
  *  -Pi < A <= Pi.  PI is the MP value of Pi computed by a previous call to
  *  MPPI.  For extra high levels of precision, use MPANGX.  The last word of
  *  the result is not reliable.  Debug output starts with MPIDB = 5.
  *  <br>
  *  Dimension a(mpnw+2), pi(mpnw), x(mpnw), y(mpnw).
  *  <br>
  *  The Taylor series for Sin converges much more slowly than that of Arcsin.
  *  Thus this routine does not employ Taylor series, but instead computes
  *  Arccos or Arcsin by solving Cos (a) = x or Sin (a) = y using one of the
  *  following Newton iterations, both of which converge to a:
  *  <pre>
  *    z_{k+1} = z_k - [x - Cos (z_k)] / Sin (z_k)
  *    z_{k+1} = z_k + [y - Sin (z_k)] / Cos (z_k)
  *  </pre>
  *  The first is selected if Abs (x) <= Abs (y); otherwise the second is used.
  *  These iterations are performed with a maximum precision level MPNW that
  *  is dynamically changed, approximately doubling with each iteration.
  *  See the comment about the parameter NIT in MPDIVX.
  *  @see #mppi
  *  @see mp_complex#mpangx
  *  @see #mpidvx
  */
  static void mpang (const mp& x, const mp& y, const mp& pi, mp& a, int mpnw);
  
  /**
  *  This computes the hyperbolic cosine and sine of the MP number A and
  *  returns the two MP results in X and Y, respectively.  
  *  
  *  AL2 is the MP value of Log (10) computed by a previous call to MPLOG.  
  *  For extra high levels of precision, use MPCSHX.  The last word of the result 
  *  is not reliable. Debug output starts with MPIDB = 5.
  *  <br>
  *  Dimension a(mpnw), al2(mpnw), x(mpnw+2), y(mpnw+2).
  *  <br>
  *  @see #mplog
  *  @see #mpcshx
  */
  static void mpcssh (const mp& a, const mp& al2, mp& x, mp& y, int mpnw);
  
  /**
  *  This computes the cosine and sine of the MP number A and returns the two MP
  *  results in X and Y, respectively. 
  *
  *  PI is the MP value of Pi computed by a previous  to MPPI.  
  *  For extra high levels of precision, use MPCSSX.
  *  The last word of the result is not reliable.  Debug output starts with
  *  MPIDB = 6.
  *  <br>
  *  Dimension a(mpnw), pi(mpnw), x(mpnw+2), y(mpnw+2).
  *  <br>
  *  This routine uses the conventional Taylor's series for Sin (s):
  *  <pre>
  *  Sin (s) =  s - s^3 / 3! + s^5 / 5! - s^7 / 7! ...
  *  </pre>
  *  where s = t - a * pi / 2 - b * pi / 16 and the integers a and b are chosen
  *  to minimize the absolute value of s.  We can then compute
  *  <pre>
  *  Sin (t) = Sin (s + a * pi / 2 + b * pi / 16)
  *  Cos (t) = Cos (s + a * pi / 2 + b * pi / 16)
  *  </pre>
  *  by applying elementary trig identities for sums.  The sine and cosine of
  *  b * pi / 16 are of the form 1/2 * Sqrt {2 +- Sqrt [2 +- Sqrt(2)]}.
  *  Reducing t in this manner insures that -Pi / 32 < s <= Pi / 32, which
  *  accelerates convergence in the above series.
  *  @see #mppi
  *  @see mp_complex#mpcssx
  */
  static void mpcssn (const mp& a, const mp& pi, mp& x, mp& y, int mpnw);
  
  /**
  *  This computes the exponential function of the MP number A and returns the
  *  MP result in B.  
  *
  *  AL2 is the MP value of Log(2) produced by a prior to MPLOG.
  *  For extra high levels of precision, use MPEXPX.  The last
  *  word of the result is not reliable.  Debug output starts with MPIDB = 7.
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2), al2(mpnw).
  *  <br>
  *  This routine uses a modification of the Taylor's series for Exp (t):
  *  <pre>
  *  Exp (t) =  (1 + r + r^2 / 2! + r^3 / 3! + r^4 / 4! ...) ^ q * 2 ^ n
  *  </pre>
  *  where q = 256, r = t' / q, t' = t - n Log(2) and where n is chosen so
  *  that -0.5 Log(2) < t' <= 0.5 Log(2).  Reducing t mod Log(2) and
  *  dividing by 256 insures that -0.001 < r <= 0.001, which accelerates
  *  convergence in the above series.
  *  @see #mpexpx
  *  @see #mplog
  */
  static void mpexp (const mp& a, const mp& al2, mp& b, int mpnw);
  
  /**
  *  This computes the natural logarithm of the MP number A and returns the MP
  *  result in B.  
  *
  *  AL2 is the MP value of Log(2) produced by a prior  to MPLOG. 
  *  For extra high levels of precision, use MPLOGX.  The last word of
  *  the result is not reliable.  Debug output starts with MPIDB = 6.
  *  <br>
  *  Dimension a(mpnw), al2(mpnw), b(mpnw+2).
  *  <br>
  *  The Taylor series for Log converges much more slowly than that of Exp.
  *  Thus this routine does not employ Taylor series, but instead computes
  *  logarithms by solving Exp (b) = a using the following Newton iteration,
  *  which converges to b:
  *  <pre>
  *    x_{k+1} = x_k + [a - Exp (x_k)] / Exp (x_k)
  *  </pre>
  *  These iterations are performed with a maximum precision level MPNW that
  *  is dynamically changed, approximately doubling with each iteration.
  *  See the comment about the parameter NIT in MPDIVX.
  *  @see #mpdivx
  *  @see #mplogx
  */
  static void mplog (const mp& a, const mp& al2, mp& b, int mpnw);
  
  /**
  *  This computes Pi to available precision (MPNW mantissa words).  
  *
  *  For extra high levels of precision, use MPPIX.  The last word of 
  *  the result is not reliable.  Debug output starts with MPIDB = 7.
  *  <br>
  *  Dimension pi(mpnw+2).
  *  <br>
  *  The algorithm that is used for computing Pi, which is due to Salamin
  *  and Brent, is as follows:
  *  <pre>
  *  Set  A_0 = 1,  B_0 = 1/Sqrt(2)  and  D_0 = Sqrt(2) - 1/2.
  *  </pre>
  *  then from k = 1 iterate the following operations:
  *  <pre>
  *  A_k = 0.5 * (A_{k-1} + B_{k-1})
  *  B_k = Sqrt (A_{k-1} * B_{k-1})
  *  D_k = D_{k-1} - 2^k * (A_k - B_k) ^ 2
  *  </pre>
  *  Then  P_k = (A_k + B_k) ^ 2 / D_k  converges quadratically to Pi.
  *  In other words, each iteration approximately doubles the number of correct
  *  digits, providing all iterations are done with the maximum precision.
  *  @see #mppix
  */
  static void mppi(mp& pi, int mpnw);


//************************ FFT CORE ROUTINES  *******************************
  
  /**
  *  This performs an N-point complex-to-real FFT, where N = 2^M.  X is the
  *  double complex input array, and Y is the double output array.
  * 
  *  The array X is used as a scratch array in MPFFT1, and so is overwritten.
  *  X must be dimensioned with N/2+N1*NSP1+1 DC cells, and Y with N DP cells, 
  *  where N = 2^M and N1 = 2^int(M/2).   This Dimension requirement for X is 
  *  somewhat greater than shown in the Dimension statement below, because 
  *  MPFFT1, which is called by this routine, requires more.  IS is the SIGN of
  *  the transform.  Before calling MPFFTCR, the UU1 and UU2 arrays must be 
  *  initialized by calling MPINIX.  This routine is not intended to be called 
  *  directly by the user.
  *  @see #init_mpuu1
  *  @see #mpfft1
  */
  static void mpfftcr (int is, int m, int n, DComplex x[], double y[]);
  
  /**
  *  This performs an N-point real-to-complex FFT, where N = 2^M.  X is the
  *  double input array, and Y is the double complex output array.
  *
  *  X must be dimensioned with N DP cells, and Y with N/2+N1*NSP1+1 DC cells,
  *  where N = 2^M and N1 = 2^int(M/2).  This Dimension requirement for Y is 
  *  somewhat greater than that shown in the Dimension statement below, because
  *  MPFFT1, which is called by this routine, requires more.  IS is the SIGN of
  *  the transform.  Before calling MPFFTRC, the UU1 and UU2 arrays must be 
  *  initialized by calling MPINIX.  This routine is not intended to be called 
  *  directly by the user.
  *  @see #init_mpuu1
  */
  static void mpfftrc (int is, int m, int n, const double x[], DComplex y[]);
  
    /**
    *  This routine performs a complex-to-complex FFT.  
    *  
    *  IS is the SIGN of the transform, 
    *  N = 2^M is the size of the transform.  N1 = 2^M1 and N2 = 2^M2,
    *  where M1 and M2 are defined as below.  X is the input and output array,
    *  and Y is a scratch array.  X must have at N, and Y at least N + N1*MPNSP1,
    *  double complex cells.  The arrays MPUU1 and MPUU2 must have been 
    *  initialized by calling MPINIX.  This routine is not intended to be called 
    *  directly by the user.
    *  <br>
    *  This employs the two-pass variant of the "four-step" FFT.  See the
    *  article by David H. Bailey in J. of Supercomputing, March 1990, p. 23-35.
    *  @see #init_mpuu1
  */
  static void mpfft1 (int is, int m, int n1, int n2, DComplex x[], DComplex y[]);
  
  /**
  *  This performs NS simultaneous N-point complex-to-complex FFTs, where
  *  N = 2^M.  
  *
  *  X is the input and output array, UU1 is the root of unity array,
  *  and Y is a scratch array.  X, Y and UU1 are double complex.   This routine
  *  is not intended to be called directly by the user.
  */
  static void mpfft2 (int is, int ns, int m, int n, DComplex **x, DComplex **y);
  
  /**
  *  This performs the L-th iteration of the second variant of the Stockham FFT
  *  on the NS vectors in X.  
  *
  *  Y is a scratch array, and UU1 is the root of unity array.  
  *  X, Y and UU1 are double complex.  This routine is not
  *  intended to be called directly by the user.
  */
  static void mpfft3 (int is, int l, int ns, int m, int n, DComplex **x, DComplex **y);
  
  /**
  * This computes the linear convolution of the N-long inputs A and B.  
  *
  * |IQ| is the number of arguments (i.e., if IQ = 1, then B is ignored).  
  * If IQ is negative (and N < 64) then only the second half of the result vector is
  * required (i.e. this is a  by itself -- see below). NSQ = int(sqrt(3*N))
  * is an input required for the Dimension of DC1 and DC2 (see below). 
  * <br>
  * This routine employs an advanced FFT-based scheme, except for small n.
  * This routine is not intended to be called directly by the user.
  * <br> 
  *  Two machine-dependent parameters are set in this routine:
  *  <br>ERM = Maximum tolerated FFT roundoff error.  On IEEE systems ERM =
  *    0.438D0.  It is not necessary to specify ERM for modest levels of
  *    precision -- see comments below.
  *  <br>MBT = Number of mantissa bits in double data.  MBT = 53 on
  *    IEEE systems, and MBT = 48 (i.e. single precision) on Crays.
  *    It is not necessary to specify MBT for modest levels of precision.
  */
  static void mplconv (int iq, int n, int nsq, double a[], double b[], double c[]);
    
  
  //************************ADVANCE ARITHMETIC ROUTINES ************************
  
  /**
  *  This computes the cube root of the MP number A and returns the MP result
  *  in B.  
  *  
  *  Before calling MPCBRX, the arrays UU1 and UU2 must be initialized by
  *  calling MPINIX.  For modest levels of precision, use MPCBRT.  Debug output
  *  starts with MPIDB = 6.
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2).
  *  <br>
  *  This routine uses basically the same Newton iteration algorithm as MPCBRT.
  *  In fact, this routine calls MPCBRT to obtain an initial approximation.
  *  See the comment about the parameter NIT in MPDIVX.
  *  @see #mpcbrt
  *  @see #mpdivx
  *  @see #mpinix
  */
  static void mpcbrx (const mp& a, mp& b, int mpnw);
  
  /**
  *  This divides the MP number A by the MP number B and returns the MP result
  *  in c.  
  *
  *  Before calling MPDIVX, the arrays UU1 and UU2 must be initialized by
  *  calling MPINIX.  For modest levels of precision, use MPDIV.  Debug output 
  *  starts with MPIDB = 7.
  *  <br>
  *  Dimension a(mpnw), b(mpnw), c(mpnw+2). 
  *  <br>
  *  This subroutine employs the following Newton-Raphson iteration, which
  *  converges to 1 / B:
  *  <pre>
  *   X_{k+1} = X_k + (1 - X_k * B) * X_k
  *  </pre>
  *  where the muliplication () * X_k is performed with only half of the
  *  normal level of precision.  These iterations are performed with a
  *  maximum precision level MPNW that is dynamically changed, doubling with
  *  each iteration.  The final iteration is performed as follows (this is
  *  due to A. Karp):
  *  <pre>
  *   A / B = (A * X_n) + [A - (A * X_n) * B] * X_n  (approx.)
  *  </pre>
  *  where the multiplications A * X_n and [] * X_n are performed with only
  *  half of the final level of precision.
  *  <br>
  *  One difficulty with this procedure is that errors often accumulate in the
  *  trailing mantissa words.  This error can be controlled by repeating one of
  *  the iterations.  The iteration that is repeated is controlled by setting
  *  the parameter NIT below:  If NIT = 0, the last iteration is repeated (this
  *  is most effective but most expensive).  If NIT = 1, { the next-to-last
  *  iteration is repeated, etc.
  *  @see #mpdiv
  *  @see #mpinix
  */
  static void mpdivx (const mp& a, const mp& b, mp& c, int mpnw);
  
  /**
  *  This routine multiplies MP numbers A and B to yield the MP product c.
  *
  *  Before calling MPMULX, the arrays UU1 and UU2 must be initialized by
  *  calling MPINIX.  For modest levels of precision, use MPMUL.  Debug output 
  *  starts with MPIDB = 8.
  *  <br>
  *  Dimension a(mpnw+), b(mpnw), c(mpnw+2).
  *  <br>
  *  This routine returns up to MPNW mantissa words of the product.  If the
  *  complete double-long product of A and B is desired (for example in large
  *  integer applications), then MPNW must be at least as large as the sum of
  *  the mantissa lengths of A and B.  In other words, if the precision levels
  *  of A and B are both 256 words, { MPNW must be at least 512 words to
  *  obtain the complete double-long product in c.
  *  <br>
  *  This subroutine uses an advanced technique involving the fast Fourier
  *  transform (FFT).  For high precision it is significantly faster than the
  *  conventional scheme used in MPMUL.
  *  @see #mpmul
  *  @see #mpinix
  */
  static void mpmulx(const mp& a, const mp&b, mp& c, int mpnw);
  
  /**
  *  This computes the N-th power of the MP number A and returns the MP result
  *  in B. 
  *
  *  When N is zero, 1 is returned.  When N is negative, the reciprocal
  *  of A ^ |N| is returned.  Before calling MPNPWX, the arrays UU1 and UU2
  *  must be initialized by calling MPINIX.  For modest levels of precision, use
  *  MPNPWR.  Debug output starts with MPIDB = 6.
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2).
  *  <br>
  *  This routine employs the binary method for exponentiation.
  *  @see #mpnpwr
  *  @see #mpinix
  */
  static void mpnpwx (const mp& a, int n, mp& b, int mpnw);
  
  /**
  *  This computes the N-th root of the MP number A and returns the MP result
  *  in B.
  *
  *  N must be at least one and must not exceed 2 ^ 30.  Before calling
  *  MPNRTX, the arrays UU1 and UU2 must be initialized by calling MPINIX.  For 
  *  modest levels of precision, use MPNRT.  Debug output starts with MPIDB = 6.
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2).
  *  <br>
  *  This routine uses basically the same Newton iteration algorithm as MPNRT.
  *  In fact, this routine calls MPNRT to obtain an initial approximation.
  *  See the comment about the parameter NIT in MPDIVX.
  *  @see #mpdivx.
  *  @see #mpnrt
  *  @see #mpinix
  */
  static void mpnrtx (const mp& a, int n, mp& b, int mpnw);
  
  /**
  *  This computes the square root of the MP number A and returns the MP result
  *  in B.
  *
  *  Before calling MPSQRX, the arrays UU1 and UU2 must be initialized by
  *  calling MPINIX.  For modest levels of precision, use MPSQRT.  Debug output
  *  starts with MPIDB = 6.
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2).
  *  <br>
  *  This routine uses basically the same Newton iteration algorithm as MPSQRT.
  *  In fact, this routine calls MPSQRT to obtain an initial approximation.
  *  See the comment about the parameter NIT in MPDIVX.
  *  @see #mpsqrt
  *  @see #mpdivx
  *  @see #mpinix
  */
  static void mpsqrx (const mp& a, mp& b, int mpnw);
  
  /**
  *  This routine squares the MP number A to yield the MP product B.
  *
  *  Before calling MPSQX, the arrays UU1 and UU2 must be initialized by calling
  *  MPINIX.  For modest levels of precision, use MPMUL.  MPNW should be a power
  *  of two.  Debug output starts with MPIDB = 8.
  *  <br>
  *  Dimension a(mpnw), b(mpnw+2).
  *  <br>
  *  This subroutine uses the same FFT technique as MPMULX.  It is faster
  *  because only one forward FFT has to be computed.  See the comments in
  *  MPMULX about obtaining the complete double-long result.
  *  @see #mpinix
  *  @see #mpmul
  */
  static void mpsqx(const mp& a, mp& b, int mpnw);
  
  //************* ADVANCE ALGEBRAIC & TRANSCEDENTAL ROUTINES ******************
  
  /**
  *  This computes the hyperbolic cosine and sine of the MP number A and
  *  returns the two MP results in X and Y, respectively. 
  *  
  *  PI is the MP value of Pi computed by a previous  to MPPI or MPPIX.  
  *  AL2 is the MP value of Log (10) computed by a previous  to MPLOG or MPLOGX.  
  *  Before calling MPCSHX, the arrays UU1 and Uu2 must be initialized by calling
  *  MPINIX.  For modest levels of precision, use MPCSSH.  The last word 
  *  of the result is not reliable.  Debug output starts with MPIDB = 5.
  *  <br>
  *  Dimension a(mpnw), al2(mpnw), x(mpnw+2), y(mpnw+2), pi(mpnw).
  *  @see #mppi
  *  @see #mppix
  *  @see #mplog
  *  @see #mplogx
  *  @see #mpcssh
  *  @see #mpinix
  */
  static void mpcshx (const mp& a, const mp& pi, const mp& al2, 
    mp& x, mp& y, int mpnw);
  
    /**
    *  This computes the exponential function of the MP number A and returns the
    *  MP result in B.  
    *
    *  PI is the MP value of Pi produced by a prior  to MPPI or MPPIX.  
    *  AL2 is the MP value of Log(2) produced by a prior  to
    *  MPLOG  or MPLOGX.  Before calling MPEXPX, the arrays UU1 and UU2 must be
    *  initialized by calling MPINIX.  For modest levels of precision, use MPEXP.
    *  The last word of the result is not reliable.  Debug output starts 
    *  with MPIDB = 5.  
    *  <br>
    *  Dimension a(mpnw), al2(mpnw), b(mpnw+2), pi(mpnw).
    *  <br>
    *  This routine uses the Newton iteration
    *  <pre>
    *    b_{k+1} = b_k [a + 1 - log b_k]
    *  </pre>
    *  with a dynamically changing level of precision.  Logs are performed using
    *  MPLOGX.  See the comment about the parameter NIT in MPDIVX.
    *  @see #mppi
    *  @see #mppix
    *  @see #mplog
    *  @see #mplogx
    *  @see #mpinix
    *  @see #mpepx
  */
  static void mpexpx (const mp& a, const mp& pi, const mp& al2, mp& b, int mpnw);
  
  /**
  *  This computes the natural logarithm of the MP number A and returns the MP
  *  result in B.  
  *
  *  PI is the MP value of Pi produced by a prior  to MPPI or MPPIX.  
  *  AL2 is the MP value of Log(2) produced by a prior  to MPLOG
  *  or MPLOGX.  Before calling MPLOGX, the arrays UU1 and UU2 must be
  *  initialized by calling MPINIX.  For modest levels of precision, use MPLOG.
  *  The last word of the result is not reliable.  Debug output starts 
  *  with MPIDB = 6.
  *  <br>
  *  Dimension al2(mpnw), pi(mpnw), a(mpnw), b(mpnw+2).
  *  <br>
  *  This uses the following algorithm, which is due to Salamin.  If a is
  *  extremely close to 1, use a Taylor series.  Otherwise select n such that
  *  z = x 2^n is at least 2^m, where m is the number of bits of desired
  *  precision in the result.  Then
  *  <pre>
  *  Log(x) = Pi / [2 AGM (1, 4/x)]
  *  </pre>
  *  @see #mppi
  *  @see #mppix
  *  @see #mplog
  *  @see #mplogx
  *  @see #mpinix
  */
  static void mplogx (const mp& a, const mp& pi, const mp& al2, mp& b, int mpnw);
  
  /**
  *  This computes Pi to available precision (MPNW mantissa words). 
  *
  *  Before calling MPPIX, the arrays UU1 and UU2 must be initialized by calling
  *  MPINIX.  For modest levels of precision, use MPPI.  MPNW should be a power
  *  of two.  The last word of the result is not reliable.  Debug
  *  output starts with MPIDB = 7.
  *  <br>
  *  Dimension pi(mpnw+2).
  *  <br>
  *  This routine uses basically the same algorithm as MPPI.
  *  @see #mppi
  *  @see #mpinix
  */
  static void mppix (mp& pi, int mpnw);
  
  
  //******************** ADVANCE IO / DEBUG ROUTINES ***************************
  
  /**
  *  Converts the MP number A into character form in the char array B.
  *
  *  N (an output parameter) is the length of the output.  In other words, B is 
  *  contained in B[0], ..., B[N-1].  The format is analogous to the Fortran
  *  exponential format (E format), except that the exponent is placed first.
  *  Before calling MPOUTX, the arrays UU1 and UU2 must be initialized by
  *  calling MPINIX.  For modest levels of precision, use MPOUTC.  Debug output
  *  starts with MPIDB = 7.
  *  <br>
  *  Dimension a(mpnw), b(7.225 * MPNW + 30 + 1).
  *  @see #mpinix
  *  @see #mpoutc
  */
  static void mpoutx (const mp& a, char* b, int& n, int mpnw);

 /**
  *  This routine converts the char array  A, which
  *  represents a multiprecision number in C++ style, i.e.
  *  "1234567890" or "1.23456789e-21", into standard MP binary format.
  *
  *  This routine is not intended to be called directly by the user.
  *  <br>
  *  Dimension b(mpnw+2)
  */
  static void mpdexc (const char a[], int l, mp& b, int mpnw);

private:
 
  /**
  *  Assign 0 to mp number in
  */
  inline static void zero(mp& in)
  {in.nw=0; in.sign=1; in.exponent=0;}
  
  /**  
  *  Helper function for mpinpc.
  *  @see #mpinpc
  */
  static void mpinpcExit(); 
  
  /**
  *  This performs rounding and truncation of the MP number A.  
  *  
  *  It is called by MPNORM, and also by other subroutines when the precision level is
  *  reduced by one.  It is not intended to be directly called by the user.
  *  <br>
  *  Dimension a(mpnw+2).
  *  <br>
  *  The parameter AMX is the absolute value of the largest exponent word
  *  allowed for MP numbers.
  *  @see #mpnorm
  */
  static void mproun(mp& a, int mpnw);
  
  /**
  *  This converts the MP number in array D of MPCOM4 to the standard
  *  normalized form in A.  
  *
  *  The MP routines often leave negative numbers or values exceeding the 
  *  radix MPBDX in result arrays, and this fixes them.
  *  MPNORM assumes that two extra mantissa words are input at the end of D.
  *  This reduces precision loss when it is necessary to shift the result to
  *  the left.  This routine is not intended to be called directly by the user.
  *  Debug output starts with MPIDB = 10. 
  *  <br>
  *  Dimension d(mpnw+4), a(mpnw+2).
  *  d is a double array with David's mp number format. 
  */
  static void mpnorm(double d[], mp& a, int mpnw);
  
  /**
  *  This performs the arithmetic-geometric mean (AGM) iterations. 
  *
  *  This routine is called by MPLOGX.  It is not intended to be called directly by the user.
  *  <br>
  *  Dimension a(mpnw+2), b(mpnw+2).
  *  @see #mplogx
  */
  static void mpagmx (mp& a, mp& b, int mpnw);

 
 

 /********************** FRIENDS ************************/
  friend class mp_complex;
  #include "mpint_friends.h"
  #include "mpreal_friends.h"
  #include "mpcomplex_friends.h"
  
};

#endif
