#ifndef MP_COMPLEX_H
#define MP_COMPLEX_H

class mp_int;
class mp_real;
#include "mpreal.h"

/**  
* This class represents multi-precision complex.
*
* @author Herman Harjono
* @version Sep 21, 1998.
*/ 
class mp_complex: public mpfun
{
  #include "mpcomplex_friends.h"
  friend class mpreal;
  #include "mpreal_friends.h"

 public:

 inline const mp* getReal() const {return r;}
  
  /**
  *  Creates a mp_complex (initialized to zero). 
  *  @param precision The precision level of the object, in digits. I
  *  @see mp#mpipl
  */
  mp_complex(const mpprecision& precision=mp::mpipl);
  
  /**
  * Destructor.
  * Perform clean up unless the private data members point to external objects.
  */
  inline ~mp_complex() { if (clean) {delete r; delete i;}}
  
  /**
  *  Creates a mp_complex from int. 
  *  @param precision The precision level of the object, in digits. 
  *  @see mp#mpipl
  */
  mp_complex(int, const mpprecision& precision=mp::mpipl);
  
  /**
  *  Creates a mp_complex from double. 
  *  @param precision The precision level of the object, in digits. 
  *  @see mp#mpipl
  */
  mp_complex(double, const mpprecision& precision=mp::mpipl);
  
  /**
  *  Creates a mp_complex from DComplex. 
  *  @param precision The precision level of the object, in digits. 
  *  @see mp#mpipl
  */
  mp_complex(const DComplex& dc, const mpprecision& precision=mp::mpipl);
  
  /**
  *  Creates a mp_complex from char*. 
  *  @param precision The precision level of the object, in digits. 
  *  @see mp#mpipl
  */
  mp_complex(char *, const mpprecision& precision=mp::mpipl);
  
  
  /**
  *  Creates a mp_real from mp. Inherits the mp object's precision.
  */
  mp_complex(const mp&);
  
  /**
  *  Copy constructor. Calls mpceq.
  */
  mp_complex(const mp_complex&);
  
  /**
  *  Creates a mp_complex from int. 
  *  @param precision The precision level of the object, in digits. 
  *  @see mp#mpipl
  */
  mp_complex(int, int, const mpprecision& precision=mp::mpipl);
  
  /**
  *  Creates a mp_complex from float/double. 
  *  @param precision The precision level of the object, in digits. 
  *  @see mp#mpipl
  */
  mp_complex(double, double, const mpprecision& precision=mp::mpipl);
  
  /**
  *  Creates a mp_complex from char*. 
  *  @see mp#mpipl
  */
  mp_complex(char *, char *, const mpprecision& precision=mp::mpipl);
  
  /**
  *  Creates a mp_real from mp. Inherits the mp objects precisions.
  */
  mp_complex(const mp&, const mp&);
  
  /**
  *  Assignment operator.
  */
  mp_complex& operator=(const int&);
  
  /**
  *  Assignment operator.
  */
  mp_complex& operator=(const double&);
  
  /**
  *  Assignment operator.
  */
  mp_complex& operator=(const mp&);
  
  /**
  *  Assignment operator.
  */
  mp_complex& operator=(const mp_complex&);
  
  /**
  *  Calls operator+ and operator=
  */
  mp_complex& operator+=(const mp_complex& za)
  {*this = (*this)+za; return *this;}
  
  /**
  *  Calls operator- and operator=
  */
  mp_complex& operator-=(const mp_complex& za)
  {*this = (*this)-za; return *this;}
  
  /**
  *  Calls operator* and operator=
  */
  mp_complex& operator*=(const mp_complex& za)
  {*this = (*this)*za; return *this;}
  
  /**
  *  Calls operator/ and operator=
  */
  mp_complex& operator/=(const mp_complex& za)
  {*this = (*this)/za; return *this;}

private:
/** 
 *  Real part.
  */
  mp* r;
  
  /**
  *  Imaginary part.
  */
  mp* i;
  
  /**
  *  Flag to indicate whether destructor perform clean up or not.
  */
  bool clean;
  
  /**
  *  Returns a pointer to a mp_complex whose value is zero.
  *
  *  It's user responsibility to delete the created object.
  *  @param maxnw the number of words allocated for the real and imaginary part.
  *  @return a pointer to the newly created object.
  */
  static mp_complex* createComplex(int maxnw);  
  
  
#if DEBUG == 2
  /**
  *  Helper function for debugging.
  */
  inline static void debug(char* name, const mp_complex& in) 
  { mp::debug(name, *in.r); mp::debug(name, *in.i);} 
#endif
  
  /**
  * Rounding a complex number.
  */
  inline static void mpcroun(mp_complex& in, int mpnw)
  { mp::mproun(*in.r, mpnw); mp::mproun(*in.i, mpnw);}
  
  /**
  * Initialize the complex number to zero.
  */
  inline static void zero(mp_complex& in)
  {mp::zero(*in.r); mp::zero(*in.i);}
  
  /**
  *  This performs the arithmetic-geometric mean (AGM) iterations.  
  *
  *  This routine is called by MPANGX.  It is not intended to be called directly by the user.
  *  <br>
  *  Dimension a(2*mpnw+4), b(2*mpnw+4).
  *  @see #mpangx
  */
  static void mpcagx (mp_complex& a, mp_complex& b, int mpnw);
  
  /**
  *  Constructor helper function
  */
  inline int constHelper(const mpprecision& precision);
  
protected:
  
/**
*  Returns a pointer to a mp_complex.
* 
*  It's user responsibility to delete created the object.
*  Note that the arguments are not const-aguments. 
*  The created mp_complex object will use those arguments as its internal 
*  data structure. The destructor won't delete those data structure.
  */
  static mp_complex* createComplex(mp& real, mp& imaginary);
  
  //************************* ARITHMETIC ROUTINES ****************************
  
  /**
  *  This computes the sum of the MPC numbers A and B and returns the MPC
  *  result in C.  
  *
  *  L is the offset between real and i parts in A, B, and c.
  *  Debug output starts with MPIDB = 9.
  *  <br>
  *  Dimension a(2*l), b(2*l), c(2*l).L must be at least MPNW + 2.
  */
  static void mpcadd (const mp_complex& a, const mp_complex& b, 
    mp_complex& c, int mpnw);
  
    /**
    *  This routine divides the MP complex numbers A and B to yield the MPC
    *  quotient c.  
    *
    *  For extra high levels of precision, use MPCDVX. 
    *  The last word of the result is not reliable.  
    *  Debug output starts with MPIDB = 7
    *  <br>
    *  Dimension a(2*l), b(2*l), c(2*l). L must be at least MPNW + 2. 
    *  <br>
    *  This routine employs the formula described in MPCMUL to save multiprecision
    *  multiplications.
    *  @see #mpcdvx
    *  @see #mpcmul
  */
  static void mpcdiv (const mp_complex& a, const mp_complex& b, 
    mp_complex& c, int mpnw);
  
    /**
    *  This sets the MPC number B equal to the MPC number A. 
    *
    *  Debug output starts with MPIDB = 10.
    *  <br>
    *  Dimension a(2*l), b(2*l).
  */
  static void mpceq (const mp_complex& a, mp_complex& b, int mpnw);
  
  /**
  *  This routine multiplies the MP complex numbers A and B to yield the MPC
  *  product c. 
  *
  *  For extra high levels of precision, use MPCMLX.  
  *  The last word of the result is not reliable.  
  *  Debug output starts with MPIDB = 7.
  *  <br>
  *  Dimension a(2*l), b(2*l), c(2*l). L must be at least MPNW + 2.
  *  <br>
  *  This routine employs the formula:
  *  <pre>
  *  (a_1 + a_2 i) (b_1 + b_2 i)  =  [a_1 b_1 - a_2 b_2]  +
  *   [(a_1 + a_2) (b_1 + b_2) - (a_1 b_1 + a_2 b_2)] i
  *  </pre>
  *  Note that this formula can be implemented with only three multiplications
  *  whereas the conventional formula requires four.
  *  @see #mpcmlx
  */
  static void mpcmul(const mp_complex& a, const mp_complex& b, 
    mp_complex& c, int mpnw);
  
    /**
    *  This computes the N-th power of the MPC number A and returns the MPC
    *  result C in B.  
    *
    *  When N is zero, 1 is returned.  When N is negative, the
    *  reciprocal of A ^ |N| is returned. For extra high levels of precision, 
    *  use MPCPWX.  Debug output starts with MPIDB = 7.
    *  <br>
    *  Dimension a(2*l), b(2*l). L should be at least MPNW + 2. 
    *  <br>
    *  This routine employs the binary method for exponentiation.
    *  @see #mpcpwx
  */
  static void mpcpwr (const mp_complex& a, int n, 
    mp_complex& b, int mpnw);
  
    /**
    *  This routine computes the complex square root of the MPC number A and 
    *  put the result in C. 
    *
    *  For extra high levels of precision, use MPCSQX.  
    *  The last word of the result is not reliable.  
    *  Debug output starts with MPIDB = 6.
    *  <br>
    *  Dimension a(2*l), b(2*l). L must be at least MPNW + 2.
    *  <br>
    *  This routine uses the following formula, where A1 and A2 are the real and
    *  imaginary parts of A, and where R = Sqrt [A1 ^ 2 + A2 ^2]:
    *  <pre>
    *  B = Sqrt [(R + A1) / 2] + I Sqrt [(R - A1) / 2]
    *  </pre>
    *  If the imaginary part of A is < 0, then the imaginary part of B is also
    *  set to be < 0.
    *  @see #mpcsqx
  */
  static void mpcsqt (const mp_complex& a, mp_complex& b, int mpnw);
  
  /**
  *  This subracts the MPC numbers A and B and returns the MPC difference in
  *  C. 
  *
  *  Debug output starts with MPIDB = 9.
  *  <br>
  *  Dimension a(2*l), b(2*l), c(2*l). L must be at least MPNW + 2.
  */
  static void mpcsub (const mp_complex& a, const mp_complex& b, 
    mp_complex& c, int mpnw);
  
  
  //**************** ADVANCE ALGEBRAIC AND TRANSCENDENTAL ROUTINES ******************
  
  /**
  *  This computes the MP angle A subtended by the MP pair (X, Y) considered as
  *  a point in the x-y plane.  
  *
  *  This is more useful than an arctan or arcsin
  *  routine, since it places the result correctly in the full circle, i.e. 
  *  -Pi < A <= Pi.  PI is the MP value of Pi computed by a previous to
  *  MPPI or MPPIX.  Before calling MPANGX, the arrays UU1 and UU2 must be
  *  initialized by calling MPINIX.  For modest levels of precision, use MPANG.
  *  The last word of the result is not reliable.  Debug output starts 
  *  with MPIDB = 6.
  *  <br>
  *  Dimension a(mpnw+2), pi(mpnw), x(mpnw), y(mpnw).
  *  <br>
  *  This routine employs a complex arithmetic version of the MPLOGX alogirthm.
  *  @see mp#mplogx
  *  @see mp#mpinix
  *  @see mp#mppi
  *  @see mp#mppix
  *  @see mp#mpang
  */
  static void mpangx (const mp& x, const mp& y, const mp& pi, mp& a, int mpnw);
  
  /**
  *  This computes the cosine and sine of the MP number A and returns the two MP
  *  results in X and Y, respectively.  
  *
  *  PI is the MP value of Pi computed by a previous  to MPPI or MPPIX. 
  *  Before calling MPCSSX, the arrays UU1 and UU2 must be initialized 
  *  by calling MPINIX.  For modest levels of precision, use MPCSSN.  
  *  The last word of the result is not reliable.
  *  Debug output starts with MPIDB = 5.
  *  <br>
  *  Dimension a(mpnw), pi(mpnw), x(mpnw+2), y(mpnw+2).
  *  <br>
  *  This routine employs a complex arithmetic version of the scheme used in
  *  MPEXPX.  See the comment about the parameter NIT in MPDIVX.
  *  @see mp#mppi
  *  @see mp#mppix
  *  @see mp#mpcssn
  *  @see mp#expx
  *  @see mp#mpdivx
  */
  static void mpcssx (const mp& a, const mp& pi, mp& x, mp& y, int mpnw);
  
  //*********************** ADVANCE ARITHMETIC ROUTINES ******************************
  
  /**
  *  This routine divides the MP complex numbers A and B to yield the MPC
  *  quotient c.  
  *
  *  Before calling MPCDVX, the arrays UU1 and UU2 must be initialized
  *  by calling MPINIX. For modest levels of precision, use MPCDIV.
  *  The last word of the result is not reliable.  
  *  Debug output starts with MPIDB = 7
  *  <br>
  *  Dimension a(2*l), b(2*l), c(2*l). L must be at least MPNW + 2. 
  *  <br>
  *  This routine employs the same scheme as MPCDIV.
  *  @see #mpcdiv
  *  @see mp#mpinix
  */
  static void mpcdvx (const mp_complex& a, const mp_complex& b, 
    mp_complex& c, int mpnw);
  
    /**
    *  This routine multiplies the MP complex numbers A and B to yield the MPC
    *  product c. 
    *
    *  Before calling MPCMLX, the arrays UU1 and UU2 must be initialized by 
    *  calling MPINIX. For modest levels of precision, use MPCMUL. 
    *  The last word of the result is not reliable.  
    *  Debug output starts with MPIDB = 7.
    *  <br>
    *  Dimension a(2*l), b(2*l), c(2*l). L must be at least MPNW + 2.
    *  <br>
    *  This routine employs the same scheme as MPCMUL.
    *  @see #mpcmul
    *  @see mp#mpinix 
  */
  static void mpcmlx (const mp_complex& a, const mp_complex& b, 
    mp_complex& c, int mpnw);
  
    /**
    *  This computes the N-th power of the MPC number A and returns the MPC
    *  result c in B.  
    *
    *  When N is zero, 1 is returned.  When N is negative, the
    *  reciprocal of A ^ |N| is returned. Before calling MPCPWX, 
    *  the arrays UU1 and UU2 must be initialized by calling MPINIX.  
    *  For modest levels of precision, use MPCPWR.  The last word of the result 
    *  is not reliable.  Debug output starts with MPIDB = 6.
    *  <br>
    *  Dimension a(2*l), b(2*l). L should be at least MPNW + 2. 
    *  <br>
    *  This routine employs the binary method for exponentiation.
    *  @see #mpcpwr
    *  @see mp#mpinix
  */
  static void mpcpwx (const mp_complex& a, int n, 
    mp_complex& b, int mpnw);
  
    /**
    *  This routine computes the complex square root of the MPC number A.  
    *
    *  For modest levels of precision, use MPCSQT.  The last
    *  word of the result is not reliable.  Debug output starts with MPIDB = 5.
    *  <br>
    *  Dimension a(2*l), b(2*l). L must be at least MPNW + 2.
    *  <br>
    *  This routine uses the same algorithm as MPCSQT.
    *  @see #mpcsqt
  */
  static void mpcsqx (const mp_complex& a, mp_complex& b, int mpnw);
  
 
  
   
};

#endif
