#include "mpdpe.h"

double* mp_dpe::init_pt()
{
  double* pt_temp = new double[64+1];
  int i;
  pt_temp[1] = 0.50;
	
  for(i = 2; i<=64; i++)
    pt_temp[i] = 0.50 * pt_temp[i-1];
  return pt_temp;
}

void mp_dpe::normalizeDPE(mp_dpe& c)
{
  while(fabs(c.a) >= mpbdx)
  {
    c.a = mprdx * c.a;
    c.n = c.n + mpnbt;
  }
  
  while (fabs(c.a) < 1.0)
  {
    c.a = mpbdx * c.a;
    c.n = c.n - mpnbt;
  }
}


void mp_dpe::dpadd (const mp_dpe& a, const mp_dpe& b, mp_dpe& c)
{
  if(MPIER != 0)
  {
    c.a = 0;
    c.n = 0;
    return;
  }
  
  if(b.a == 0.0)
  {
    c.a = a.a;
    c.n = a.n;
  }
  else if(a.a == 0.0) 
  {
    c.a = b.a;
    c.n = b.n;
  }
  else if (a.n == b.n) 
  {
    c.a = a.a + b.a;
    c.n = a.n;
  }
  else if (a.n > b.n) 
  {
    int k = a.n - b.n;
    c.n = a.n;
    if (k > 64) 
      c.a = a.a;
    else
      c.a = a.a + b.a * pt[k];
  }
  else
  {
    int k = b.n - a.n;
    c.n = b.n;
    if (k > 64) 
      c.a = b.a;
    else
      c.a = b.a + a.a * pt[k];
  }
  
  if (c.a == 0) 
  {
    c.n = 0;
    return;
  }
  
  //   normalize the result to a decent range if it is not.
  normalizeDPE(c);
}

void mp_dpe::dpdec(const mp_dpe& a, mp_dpe& b)
{
  const double xlt = 0.3010299956639812;
  
  if (a.a != 0.0) 
  {
    double t1 = xlt * a.n + log10(fabs(a.a));
    b.n = int(t1); 
    if (t1 < 0.0) b.n = b.n - 1;
    b.a = SIGN (pow(10.0, (t1 - b.n)), a.a);
  }
  else
  {
    b.a = 0.0;
    b.n = 0;
  } 
  
  return;
}

void mp_dpe::dpdiv (const mp_dpe& a, const mp_dpe& b, mp_dpe& c)
{
  if (MPIER != 0) 
  {
    c.a = 0.0;
    c.n = 0;
    return;
  }
  if (b.a == 0.0) 
  {
    if (MPKER[1] != 0) {
      cerr <<  "('*** dpdiv: divisor is zero.')" << endl;
      MPIER = 1;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //   divide a.a by b.a and subtract exponents, unless a.a is zero.
  
  if (a.a == 0.0) 
  {
    c.a = 0.0;
    c.n = 0;
    return;
  } 
  else
  {
    c.a = a.a / b.a;
    c.n = a.n - b.n;
  }
  
  //   normalize the result to a decent range if it is not.
  normalizeDPE(c);
}


void mp_dpe::dpmul(const mp_dpe& a, const mp_dpe& b, mp_dpe& c)
{
  if (MPIER != 0) {
    c.a = 0.0;
    c.n = 0;
    return;
  }
  
  //  multiply a.a by b.a and add exponents, unless either is zero.
  
  if (a.a == 0.0 || b.a == 0.0) 
  {
    c.a = 0.0;
    c.n = 0;
    return;
  } 
  else 
  {
    c.a = a.a * b.a;
    c.n = a.n + b.n;
  }
  
  //  normalize the result to a decent range if it is not.
  normalizeDPE(c);
  
  return;
}


void mp_dpe::dppwr(const mp_dpe& a, const mp_dpe& b, mp_dpe& c)
{
  const double cl2 = 1.4426950408889633;
  
  if (MPIER != 0) 
  {
    c.a = 0.0;
    c.n = 0;
    return;
  }
  
  if (a.a <= 0.0) {
    if (MPKER[2] != 0) 
    {
      cerr << "*** dppwr: argument is less than or equal to zero." << endl;
      MPIER = 2;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  if (b.a == 0.0) {
    c.a = 1.0;
    c.n = 0;
    return;
  }
  
  if (b.a == 1.0 && b.n == 0) {
    c.a = a.a;
    c.n = a.n;
    return;
  }
  
  //  compute the base 2 logarithm of a.a and multiply by b.a.
  
  mp_dpe t1;
  mp_dpe al(cl2 * log (a.a) + a.n, 0);
	
  dpmul (al, b, t1);
	
  //  check for possible overflow or underflow.
	
  if (t1.n > 6) 
  {
    if (t1.a > 0.0) 
    {
      if (MPKER[3] != 0) 
      {
        cerr << "*** dppwr: overflow" << endl;
        MPIER = 3;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    } 
    else 
    {
      c.a = 0.0;
      c.n = 0;
      return;
    }
  }
	
  //  compute 2 raised to the power b.a * log_2 (a.a).
	
  t1.a = t1.a * pow(2.0, t1.n);
  c.n = int (t1.a);
  c.a = pow(2.0, (t1.a - c.n));
	
  //  normalize the result to a decent range if it is not.
	
  normalizeDPE(c);
}

/**
*  This computes the square root of the dpe number a and places the
*  result in b.
*/
void mp_dpe::dpsqrt(const mp_dpe& a, mp_dpe& b)
{
  if (MPIER != 0) 
  {
    b.a = 0.0;
    b.n = 0;
    return;
  }
  if (a.a < 0.0) 
  {
    if (MPKER[4] != 0) 
    {
      cerr << "*** dpsqrt: argument is negative." << endl;
      MPIER = 4;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  if (a.a == 0.0) {
    b.a = 0.0;
    b.n = 0;
    return;
  }
  
  //  divide the exponent of a.a by two and then take the square root of a.a.  if
  //  a.n is not an even number, then we have to multiply a.a by 10 before taking
  //  the square root.
  
  b.n = a.n / 2;
  if (a.n == 2 * b.n) 
  {
    b.a = sqrt (a.a);
  } 
  else 
  {
    b.a = sqrt (2.0 * a.a);
    if (a.n < 0) b.n = b.n - 1;
  }
	
  //  normalize the result to a decent range if it is not.
  normalizeDPE(b);    
}

void mp_dpe::dpsub(const mp_dpe& a, const mp_dpe& b, mp_dpe& c)
{
  
  if (MPIER != 0) 
  {
    c.a = 0.0;
    c.n = 0;
    return;
  }
  
  mp_dpe bb(-b.a, b.n);
  dpadd (a, bb, c);
}

