/**
*   This is a simplified version of Dongarra's Linpack benchmark.  The main
*   program and test procedure has been simplified, but the solution routines
*   have not been altered.  Converted to DHB's Fortran-90 based multiprecision.
*
*   David H. Bailey    Aug. 8, 1994
*   Translated to C++ (and simplified) by Herman Harjono    July 9, 1998.
*/

#include "mpreal.h"

const int nx = 100;

// prototypes
void matgen(mp_real a[nx][nx], int n, mp_real b[nx], mp_real& norma);
void dgefa(mp_real a[nx][nx], int n, int ipvt[nx], int& info);
void dgesl(mp_real a[nx][nx], int n, const int ipvt[nx], mp_real b[nx], int job);
void dmxpy(int n1, mp_real y[nx], int n2, const mp_real x[nx], const mp_real m[nx][nx]);
int idamax(int n,mp_real dx[nx][nx], int k);
void dscal(int n, mp_real da, mp_real* dx);   
void daxpy(int n, mp_real da, mp_real* dx, int incx, mp_real* dy, int incy);
mp_real ddot(int n,mp_real* dx,mp_real* dy);

int main()
{
  
  const double dpeps = 1e-16;
  mp_real a[nx][nx],b[nx],x[nx],resid,residn, norma, normx;
  int i, ipvt[nx], lda = nx, n = nx, info;
  
  matgen(a,n,b,norma);
  dgefa(a,n,ipvt,info);
  dgesl(a,n,ipvt,b,0);
	
  // compute a residual to verify results.
  
  for(i=0;i<n;i++)
    x[i] = b[i];
  
  matgen(a,n,b,norma);
	
  for(i=0;i<n;i++)
    b[i] = -b[i];
  
  
  dmxpy(n,b,n,x,a);
	
  resid = 0.0;
  normx = 0.0;
  
  
  for(i=0;i<n;i++)
  {
    resid = maximum(resid, abs (b[i]));
    normx = maximum(normx, abs (x[i]));
  }
  
  residn = resid/( n*norma*normx*mp::mpeps );
  cout << "norm, residn, resid, machep =\n" 
    << residn << resid << mp::mpeps
    << "x(1), x(n) =\n" << x[0] << x[n-1];
  
  return 0;
}


void matgen(mp_real a[nx][nx], int n, mp_real b[nx], mp_real& norma)
{
  mp_real da;
  
  int i,j,init = 1325;
  norma = 0.0;
  
  for(i = 0;i<n;i++)
  {
    for(j = 0;j<n;j++)
    {
      init = (3125*init)%65536;
      a[j][i] = (init - 32768.0)/16384.0;
      da = abs (a[j][i]);
      norma = maximum(da,norma);
    }
  }
  
  for(i = 0;i<n;i++)
    b[i]=0.0;
	
  for(i = 0;i<n;i++)
  {
    for(j = 0;j<n;j++)
      b[i] += a[i][j];
  }
  
}

/**
*  dgefa factors a double precision matrix by gaussian elimination.
* 
*  <pre>
*  dgefa is usually called by dgeco, but it can be called
*  directly with a saving in time if  rcond  is not needed.
*  (time for(dgeco) = (1 + 9/n)*(time for(dgefa) .
*
*  on entry
*
*  a       double precision(lda, n)
*          the matrix to be factored.
*
*  lda     integer
*          the leading dimension of the array  a .
*
*  n       integer
*          the order of the matrix  a .
*
*  on return
*
*  a       an upper triangular matrix and the multipliers
*          which were used to obtain it.
*          the factorization can be written  a = l*u  where
*          l  is a product of permutation and unit lower
*          triangular matrices and  u  is upper triangular.
*
*  ipvt    integer(n)
*          an integer vector of pivot indices.
*
*  info    integer
*          = 0  normal value.
*          = k  if  u(k,k) == 0.0 .  this is not an error
*               condition for(this subroutine, but it does
*               indicate that dgesl or dgedi will divide by zero
*               if called.  use  rcond  in dgeco for(a reliable
*               indication of singularity.
*
*  linpack. this version dated 08/14/78 .
*  cleve moler, university of new mexico, argonne national lab.
*
*  subroutines and functions
*
*  blas daxpy,dscal,idamax
*
*  internal variables
*  </pre>
*/
void dgefa(mp_real a[nx][nx], int n, int ipvt[nx], int& info)
{
  mp_real t;
  
  int j,k,kp1,l,nm1;
  
  // gaussian elimination with partial pivoting
  
  info = 0;
  nm1 = n - 1;
  if (nm1 < 1) 
  {
    ipvt[nm1] = n;
    if (a[nm1][nm1] == 0.0) 
      info = n;
    return;
  } 
  
  for(k = 0; k<nm1; k++)
  {
    kp1 = k + 1;
    
    //  find l = pivot index
    l = idamax(n-k,a,k) + k; 
    ipvt[k] = l;
    
    //  zero pivot implies this column already triangularized   
    if (a[l][k] == 0.0)
    {
      info=k;
      continue;
    }
    
    //  interchange if necessary
    if (l != k)
    {
      t = a[l][k];
      a[l][k] = a[k][k];
      a[k][k] = t;
    }
		
    // compute multipliers
    
    t = -1.0/a[k][k];
    dscal(n-k-1,t,&(a[k+1][k])); // scale a vector by a constant
		
		
    // row elimination with column indexing
    
    for(j = kp1; j< n; j++)
    {
      t = a[l][j];
      if (l != k)
      {
        a[l][j] = a[k][j];
        a[k][j] = t;
      }
      daxpy(n-k-1,t,&(a[k+1][k]),nx , &(a[k+1][j]), nx);    
    }
  }
  
  ipvt[nm1] = n;
  if (a[nm1][nm1] == 0.0) 
    info = n;
}


/**
*  dgesl solves the double precision system
*  a * x = b  or  trans(a) * x = b
*  using the factors computed by dgeco or dgefa.
*  <pre>
*  on entry
*
*  a       double precision(lda, n)
*          the output from dgeco or dgefa.
*
*  lda     integer
*          the leading dimension of the array  a .
*
*  n       integer
*          the order of the matrix  a .
*
*  ipvt    integer(n)
*          the pivot vector from dgeco or dgefa.
*
*  b       double precision(n)
*          the right hand side vector.
*
*  job     integer
*          = 0         to solve  a*x = b ,
*          = nonzero   to solve  trans(a)*x = b  where
*                      trans(a)  is the transpose.
*
*  on return
*
*  b       the solution vector  x .
*
*  error condition
*
*  a division by zero will occur if the input factor contains a
*  zero on the diagonal.  technically this indicates singularity
*  but it is often caused by improper arguments or improper
*  setting of lda .  it will not occur if the subroutines are
*  called correctly and if dgeco has set rcond > 0.0
*  or dgefa has set info == 0 .
*
*  to compute  inverse(a) * c  where  c  is a matrix
*  with  p  columns
*  call dgeco(a,lda,n,ipvt,rcond,z)
*  if (rcond is too small) go to ...
*  do 10 j = 1, p
*    call dgesl(a,lda,n,ipvt,c(1,j),0)
*  10 continue
*
*  linpack. this version dated 08/14/78 .
*  cleve moler, university of new mexico, argonne national lab.
*
*  subroutines and functions
*
*  blas daxpy,ddot
*
*  internal variables
* </pre>
*/
void dgesl(mp_real a[nx][nx], int n, const int ipvt[nx], mp_real b[nx], int job)
{
  mp_real t;
  
  int k,kb,l,nm1;
  
  nm1 = n - 1;
  if (job == 0)
  {
    
    // job = 0 , solve  a * x = b
    //  first solve  l*y = b
    
    if (nm1 >= 1)
    {
      for(k = 0; k<nm1; k++)
      {
        l = ipvt[k];
        t = b[l];
        if (l != k)
        {
          b[l] = b[k];
          b[k] = t;
        }
        daxpy(n-k-1,t,&(a[k+1][k]),nx, &(b[k+1]), 1);
      }
    }
    
    //  now solve  u*x = y
    
    for(kb = 1;kb<= n; kb++)
    {
      k = n  - kb;
      b[k] = b[k]/a[k][k];
      t = -b[k];
      daxpy(k,t,&(a[0][k]),nx, b,1);
    }
    return;
  }
  
  //  job = nonzero, solve  trans(a) * x = b
  //  first solve  trans(u)*y = b
  
  for(k = 0; k<n; k++)
  {
    t = ddot(k,a[k],b);
    b[k] = (b[k] - t)/a[k][k];
  }
  //  now solve trans(l)*x = y
  
  if (nm1 < 1) return;
  for(kb = 1; kb<=nm1; kb++)
  {
    k = n - kb - 1;
    b[k] += ddot(n-k+1,&a[k+1][k],&b[k+1]);
    l = ipvt[k];
    if (l == k) continue;
    t = b[l];
    b[l] = b[k];
    b[k] = t;
  }
}

/**
*   constant times a vector plus a vector.
*   jack dongarra, linpack, 3/11/78.
*/
void daxpy(int n, mp_real da, mp_real* dx, int incx, mp_real* dy, int incy)
{
  if(n<=0) return;
  if (da == 0.0) return;
  
  int i;  
  for(i = 0;i<n; i++)
    dy[incy*i] += da*dx[incx*i];  
}


/**
*  forms the dot product of two vectors.
*  jack dongarra, linpack, 3/11/78.
*/
mp_real ddot(int n,mp_real* dx,mp_real* dy)
{
	mp_real result=0.0;
	
	int i;
	if(n<=0) return result;
	
	//  code for both increments equal to 1
	
	for(i = 0; i<n;i++);
  result+= dx[i]*dy[i];
	
	return result;
}


/**
*  scales a vector by a constant.
*  jack dongarra, linpack, 3/11/78.
*/
void dscal(int n, mp_real da, mp_real* dx)
{     
  int i;
  for(i = 0;i<n; i++)
    dx[nx*i]*= da;
  
  return;
}

/**
*    finds the index of element having maximum absolute value.
*    jack dongarra, linpack, 3/11/78.
*/
int idamax(int n,mp_real dx[nx][nx],int k)
{
	mp_real dmax;
	int i;
	
	if( n < 1 ) return -1;
	
	if(n == 1) return 0;
	int result=0;
	
	dmax = abs(dx[k][k]);
	
	for(i = 1;i<n; i++)
	{
		if(abs(dx[k+i][k]) > dmax)
		{
			result = i;
			dmax = abs(dx[k+i][k]);
		}
	}
	return result;
}

/**
*   purpose: multiply matrix m times vector x and add the result to vector y.
*   <pre>
*   parameters:
*
*     n1 integer, number of elements in vector y, and number of rows in
*   matrix m
*
*     y double precision(n1), vector of length n1 to which is added 
*   the product m*x
*
*     n2 integer, number of elements in vector x, and number of columns
*   in matrix m
*
*     ldm integer, leading dimension of array m
*
*     x double precision(n2), vector of length n2
*
*     m double precision(ldm,n2), matrix of n1 rows and n2 columns
*
*   simpilifed by DHB for MP demo.
*   </pre>
*/

void dmxpy(int n1, mp_real y[nx], int n2, const mp_real x[nx], const mp_real m[nx][nx])
{
  mp_real s;
  int k,j;
  
  for(k = 0; k<n1;k++)
  {
    s = 0.0;
    
    for(j = 0; j< n2; j++)
      s += m[k][j] * x[j];
    
    y[k] += s;
  }
}


