#ifndef COMPLEX_H
#define COMPLEX_H

#include <math.h>
#include <iostream.h>

class DComplex;

/**
 *  This class represents double complex type. 
 *  @author unknown.
 *  @version Sep 21 1998, by Herman Harjono
 */
class DComplex 
{
  double re, im;
  DComplex rtop() const
    { return DComplex(hypot(re, im), atan2(im, re));}
  
  DComplex ptor() const
    {return DComplex(re*cos(im), re*sin(im));}
  
 public:
  // Constructors.
  DComplex(double r=0.0, double i=0.0)
    {re = r; im = i;}
  
  DComplex(const DComplex& d)
    { re = d.re; im = d.im;}  
  
  // Assignment.
  const DComplex& operator=(const DComplex& d)
    {
      if(&d != this)
	{re = d.re; im = d.im;}
      return *this;
    }
  
  // Field extractors.
  friend double real(const DComplex& d)
    {return d.re;}
  
  friend double aimag(const DComplex& d)
    {return d.im;}
  
  // Miscellaneous mathematical functions.
  friend double abs(const DComplex& d)
    { return hypot(d.re, d.im);}
  
  friend double arg(const DComplex& d)
    { return atan2(d.im, d.re);}
  
  
  friend DComplex conjg(const DComplex& d)
    {return DComplex(d.re, -d.im);}
  
  friend  DComplex exp(const DComplex& d)
    {
      double r = exp(d.re);
      return DComplex(r*cos(d.im), r*sin(d.im));
    }
  
  friend DComplex power(const double r, const DComplex& d)
    {
      double t = pow(r, d.re);
      return DComplex(t*cos(d.im), t*sin(d.im));
    }
  
  friend DComplex power(const DComplex& d, const double r)
    {
      DComplex polar = d.rtop();
      polar.re = pow(polar.re, r);
      polar.im *= r;
      return polar.ptor();
    }
  
  friend DComplex power(const DComplex& d, const int r)
    {
      switch(r)
	{
	  
	case 0: return 1;
	case 1: return d;
	case 2: return d * d;
	default: 
	  DComplex polar = d.rtop();
	  polar.re = pow(polar.re, r);
	  polar.im *= r;
	return polar.ptor();
	}
    }
  
  
  friend DComplex sqrt(const DComplex& d)
    {
      DComplex polar = d.rtop();
      polar.re = sqrt(polar.re);
      polar.im = 0.5*polar.im;
      return polar.ptor();
    }
  
  // Overloaded arithmetic and logical operators.
  friend DComplex operator+(const DComplex& d1, const DComplex& d2)
    {
      return DComplex(d1.re+d2.re, d1.im+d2.im);
    }
  
  friend DComplex operator-(const DComplex& d)
    {
      return DComplex(-d.re, -d.im);
    }
  
  friend DComplex operator-(const DComplex& d1, const DComplex& d2)
    {
      return DComplex(d1.re-d2.re, d1.im-d2.im);
    }
  
  friend DComplex operator*(const DComplex& d1, const DComplex& d2)
    {
      return DComplex(d1.re*d2.re-d1.im*d2.im, d1.re*d2.im+d1.im*d2.re);
    }

  friend DComplex operator/(const DComplex& d1, const DComplex& d2)
    {
      double denom = d2.re*d2.re+d2.im*d2.im;
      return DComplex((d1.re*d2.re+d1.im*d2.im)/denom,
		      (d1.im*d2.re-d1.re*d2.im)/denom);
    }
  
  friend int operator==(const DComplex& d1, const DComplex& d2)
    {
      return (d1.re==d2.re && d1.im==d2.im) ? 1 : 0;
    }
  
  friend int operator!=(const DComplex& d1, const DComplex& d2)
    {
    return (d1.re!=d2.re || d1.im!=d2.im) ? 1 : 0;
    }
  
  // More overloaded operators.
  DComplex& operator+=(const DComplex& d)
    {
      re += d.re; im += d.im;
      return *this;
    }
  
  DComplex& operator-=(const DComplex& d)
    {
      re -= d.re; im -= d.im;
      return *this;
    }

  DComplex& operator*=(const DComplex& d)
    {
      double oldRe = re;
      double oldDRe = d.re; // important if this == &d
      re = re*d.re - im*d.im; 
      im = im*oldDRe + oldRe*d.im;
      return *this;
    }

  DComplex& operator/=(const DComplex& d)
  {
    double denom = d.re*d.re+d.im*d.im;
    double oldRe = re;
    double oldDRe = d.re; // important if this == &d
    re = (re*d.re + im*d.im)/denom;
    im = (im*oldDRe - oldRe*d.im)/denom;
    return *this;
  }
  
  // Stream I/O.
  friend istream& operator>>(istream& s, DComplex& d)
    {
      s >> d.re >> d.im;
      return s;
    }

  friend ostream& operator<<(ostream& s, const DComplex& d)
    {
      s << '(' << d.re
	<< ',' << d.im
	<< ")";
      return s;
    }
};
#endif

