#include "mpcomplex.h"

mp_complex* mp_complex::createComplex(int maxnw)
{ 
  mp_complex* temp = new mp_complex();
  delete temp->r; delete temp->i;
  mpsize s(maxnw);
  temp->r = new mp(s); 
  temp->i = new mp(s);
  temp->clean = true;
  return temp;
}

mp_complex* mp_complex::createComplex(mp& real, mp& imaginary)
{
  mp_complex* temp = new mp_complex();
  delete temp->r; delete temp->i;
  temp->r = &real; 
  temp->i = &imaginary;
  temp->clean = false;
  return temp;
}

int mp_complex::constHelper(const mpprecision& precision)
{
  int maxnw = int((precision.size / 7.224719896) + 4);
  if(maxnw < 8) maxnw = 8; // since at least we need 6 (such as in mpdmc), therefore 
	// it's safe to allocate at least 8.
	
  mpsize s(maxnw);
  r = new mp(s); 
  i = new mp(s); 
  clean = true;
  return maxnw;
}

mp_complex::mp_complex(const mpprecision& precision)
{
  constHelper(precision);
}

mp_complex::mp_complex(int ia, const mpprecision& precision)
{
  constHelper(precision);
  mp::mpdmc(mp_dpe(double(ia)), *r);
}

mp_complex::mp_complex(double da,const mpprecision& precision)
{
  constHelper(precision);
  mp::mpdmc(mp_dpe(da), *r);
}

mp_complex::mp_complex(const DComplex& dc, const mpprecision& precision)
{
  constHelper(precision);
  mp::mpdmc(mp_dpe(real(dc)), *r); // mpxzc
  mp::mpdmc(mp_dpe(double(aimag(dc))), *i);
}

mp_complex::mp_complex(char* aa, const mpprecision& precision)
{
  int maxnw = constHelper(precision);
  mp::mpinpc(aa, strlen(aa), *r, MIN(MPNW, maxnw-2));
}

mp_complex::mp_complex(const mp& ja)
{
  int maxnw = ja.maxnw;
  mpsize s(maxnw);
  r = new mp(s);
  i = new mp(s);
  clean = true;
  mp::mpeq(ja, *r, MPNW); // mpmzc
}

mp_complex::mp_complex(const mp_complex& in)
{
  r = new mp(mpsize(in.r->maxnw));
  i = new mp(mpsize(in.i->maxnw));
  clean = true;
  mpceq(in, *this, MPNW);
}

mp_complex::mp_complex(int ia, int ib, const mpprecision& precision)
{
  constHelper(precision);
  
  mp::mpdmc(mp_dpe(double(ia)), *r); 
  mp::mpdmc(mp_dpe(double(ib)), *i);
}

mp_complex::mp_complex(double da, double db, const mpprecision& precision)
{
  constHelper(precision);
  
  mp::mpdmc(mp_dpe(da), *r); 
  mp::mpdmc(mp_dpe(db), *i); 
}

mp_complex::mp_complex(char* aa, char* ab, const mpprecision& precision)
{
  int maxnw = constHelper(precision); 
  int mpnw = MIN(MPNW, maxnw -1);
  mp::mpinpc(aa, strlen(aa), *r, mpnw);
  mp::mpinpc(ab, strlen(ab), *i, mpnw);
}

mp_complex::mp_complex(const mp& ja, const mp& jb)
{
  r = new mp(mpsize(ja.maxnw));
  i = new mp(mpsize(jb.maxnw));
  clean = true;
  mp::mpeq(ja, *r, MPNW); 
  mp::mpeq(jb, *i, MPNW);
}

mp_complex& mp_complex::operator=(const int& ib)
{
  mp::mpdmc(mp_dpe(double(ib)), *r);
  mp::zero(*i);
  return *this;
}

mp_complex& mp_complex::operator=(const double& db)
{
  mp::mpdmc(mp_dpe(db), *r);
  mp::zero(*i);
  return *this;
}

mp_complex& mp_complex::operator=(const mp& jb)
{
  mp::mpeq(jb, *r, MIN(MPNW, r->maxnw-1));
  mp::zero(*i);
  return *this;
}

mp_complex& mp_complex::operator=(const mp_complex& zb)
{
  mpceq(zb, *this, MIN3(MPNW, r->maxnw, i->maxnw));
  return *this;
}


mp_complex operator+(const mp_complex& za, const mp_complex& zb)
{
  mp_complex  res;
  
  mp_complex::mpcadd(za, zb, res, mp::MPNW);
  return res;
}

mp_complex operator-(const mp_complex& za, const mp_complex& zb)
{
  mp_complex res;
  
  mp_complex::mpcsub(za, zb, res, mp::MPNW);
  return res;
}

mp_complex operator-(const mp_complex& za)
{
  mp_complex res;
  
  mp_complex::mpceq(za, res, mp::MPNW);
  res.r->setSign(- res.r->getSign());
  res.i->setSign(- res.i->getSign());
  return res;
}

mp_complex operator*(const mp_complex& za, const mp_complex& zb)
{
  mp_complex  res;
  
  mp_complex::mpcmlx(za, zb, res, mp::MPNW);
  return res;
}

mp_complex operator/(const mp_complex& za, const mp_complex& zb)
{
  mp_complex  res;
  
  mp_complex::mpcdvx(za, zb, res, mp::MPNW);
  return res;
}

int operator==(const mp_complex& za, const mp_complex& zb)
{
  int ic1, ic2;
  
  mp::mpcpr(*za.r, *zb.r, ic1, mp::MPNW);
  mp::mpcpr(*za.i, *zb.i, ic2, mp::MPNW);
  return (ic1 == 0 && ic2 == 0) ? 1 : 0;
}

int operator!=(const mp_complex& za, const mp_complex& zb)
{
  int ic1, ic2;
  
  mp::mpcpr(*za.r, *zb.r, ic1, mp::MPNW);
  mp::mpcpr(*za.i, *zb.i, ic2, mp::MPNW);
  return (ic1 != 0 || ic2 != 0) ? 1 : 0;
}

mp_real abs(const mp_complex& za)
{
  mp_real res, mpt1, mpt2, mpt3;
	
  mp::mpmulx(*za.r, *za.r, mpt1, mp::MPNW);
  mp::mpmulx(*za.i, *za.i, mpt2, mp::MPNW);
  mp::mpadd(mpt1, mpt2, mpt3, mp::MPNW);
  mp::mpsqrx(mpt3, res, mp::MPNW);
  return res;
}

mp_real aimag(const mp_complex& za)
{
  mp_real res;
  
  mp::mpeq(*za.i, res, mp::MPNW);
  return res;
}

mp_complex conjg(const mp_complex& za)
{
  mp_complex  res;
	
  mp_complex::mpceq(za, res, mp::MPNW);
  res.i->setSign(- za.i->getSign());
  return res;
}

DComplex complex(const mp_complex& za)
{
  DComplex res((*za.r).toDPE().value(),
    (*za.i).toDPE().value());
  return res;
}

mp_real MP_REAL(const mp_complex& za)
{
  mp_real res;
  mp::mpeq(*za.r, res, mp::MPNW);
  return res;
}

mp_complex sqrt(const mp_complex& za)
{
  mp_complex res;
  
  mp_complex::mpcsqt(za, res, mp::MPNW);
  return res;
}


mp_complex power(const mp_complex& za, const int ib)
{
  mp_complex  res;
	
  mp_complex::mpcpwx(za, ib, res, mp::MPNW);
  return res;
}

istream& operator>>(istream& s, mp_complex& za)
  { operator>>(s, *za.r); return operator>>(s, *za.i);}

ostream& operator<<(ostream& s, const mp_complex& za)
  { operator<<(s, *za.r); return operator<<(s, *za.i);}

double dble(const mp_complex& za)
  {return dble(*za.r);}

int integer(const mp_complex& za)
  {return integer(*za.r);}
  









