#include "mpcomplex.h"

void mp_complex::mpangx (const mp& x, const mp& y, const mp& pi, mp& a, int mpnw)
{
  const double cpi = 3.141592653589793;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    mp::zero(a);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 6) 
  {
    mp::mpdeb ("MPANGX I ", x, mpnw);
    mp::mpdeb ("MPANGX I ", y, mpnw);
  }
#endif
  
  int ix = x.sign;
  int nx = MIN (x.nw, mpnw);
  int iy = y.sign;
  int ny = MIN (y.nw, mpnw);
  
  int ncr = int(pow(2, MPMCR));
  
  //  Check if precision level is too low to justify the advanced routine.
  
  if (mpnw <= ncr) 
  {
    mp::mpang (x, y, pi, a, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      mp::mpdeb ("MPANGX O ", a, mpnw);
#endif
    return;
  }
  
  //  Check if both X and Y are zero.
  
  if (nx == 0 && ny == 0) 
  {
    if (MPKER[9] != 0) 
    {
      cerr << "*** MPANGX: Both arguments are zero.\n";
      MPIER = 9;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check if Pi has been precomputed.
  mp_dpe t1;
  mp::mpmdc (pi, t1);
  if (t1.n != 0 || fabs (t1.a - cpi) > mprx2) 
  {
    if (MPKER[10] != 0) 
    {
      cerr << "*** MPANG: PI must be precomputed.\n";
      MPIER = 10;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check if one of X or Y is zero.
  if (nx == 0) 
  {
    if (iy > 0) 
      mp::mpmuld (pi, mp_dpe(0.5), a, mpnw); 
    else 
      mp::mpmuld (pi, mp_dpe(-0.5), a, mpnw); 
    
#if DEBUG == 2
    if (MPIDB >= 6)  
      mp::mpdeb ("MPANGX O ", a, mpnw);
#endif
    return;
    
  } 
  else if (ny == 0) 
  {
    if (ix > 0) 
      mp::zero(a);
    else 
      mp::mpeq (pi, a, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6)  
      mp::mpdeb ("MPANGX O ", a, mpnw);
#endif
    return;
  }
  
  //  Define scratch space.
  
  mp_complex *sk0 = createComplex(mpnw+2), 
    *sk1 = createComplex(mpnw+2), 
    *sk2 = createComplex(mpnw+2), 
    *sk3 = createComplex(mpnw+2);
  
  //  Multiply the input by a large power of two.
  mp::mpmdc (x, t1);
  int n2 = mpnbt * (mpnw / 2 + 2) - t1.n;
  mp_dpe dpe1(1.0, n2);
  mp::mpmuld (x, dpe1, *sk0->r, mpnw);
  mp::mpmuld (y, dpe1, *sk0->i, mpnw);
  
  
  //  Perform AGM iterations.
  sk1->r->nw=1; sk1->r->sign=1; sk1->r->exponent=0; (*sk1->r)[0]=1; (*sk1->r)[1]=0;
  mp::zero(*sk1->i);
  sk3->r->nw=1; sk3->r->sign=1; sk3->r->exponent=0; (*sk3->r)[0]=4; (*sk3->r)[1]=0;
  mp::zero(*sk3->i);
  
  mpcdvx (*sk3, *sk0, *sk2, mpnw);
  mpcagx (*sk1, *sk2, mpnw);
  
  //  Compute A = Imag (Pi / (2 * Z)), where Z is the limit of the complex AGM.
  
  dpe1.a = 2.0; dpe1.n = 0;
  mp::mpmuld (*sk1->r, dpe1, *sk0->r, mpnw);
  mp::mpmuld (*sk1->i, dpe1, *sk0->i, mpnw);
  mp::mpeq(pi, *sk2->r, mpnw); mp::zero(*sk2->i);
  mpcdvx (*sk2, *sk0, *sk1, mpnw);
  mp::mpeq (*sk1->i, a, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 6)  
    mp::mpdeb ("MPANGX O ", a, mpnw);
#endif
  delete sk0; delete sk1; delete sk2; delete sk3; 
  
}

void mp_complex::mpcagx (mp_complex& a, mp_complex& b, int mpnw)
{
  
  if (MPIER != 0) {
    if (MPIER == 99)  mpabrt();
    zero(a); zero(b);
    return;
  }
  
  int l1 = 0;
  mp_complex *sk0 = createComplex(mpnw+2), *sk1 = createComplex(mpnw+2);  
  int s1;
  //100  
  mp_dpe dpe1(0.50,0);
  do
  { 
    l1++;
    if (l1 == 50) 
    {
      if (MPKER[12] != 0)
      {
        cerr << "*** MPCAGX: Iteration limit exceeded.\n";
        MPIER = 12;
        delete sk0; delete sk1;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
    }
    
    s1 = sk0->r->exponent;
    mpcadd (a, b, *sk0, mpnw);
    mp::mpmuld (*sk0->r, dpe1, *sk1->r, mpnw);
    mp::mpmuld (*sk0->i, dpe1, *sk1->i, mpnw);
    mpcmlx (a, b, *sk0, mpnw);
    mpcsqx (*sk0, b, mpnw);
    mpceq (*sk1, a, mpnw);
    mp::mpsub (*a.r, *b.r, *sk0->r, mpnw);
    
  }
  //  Check for convergence.
  while (sk0->r->nw != 0. && (sk0->r->exponent < s1 || sk0->r->exponent >= -2)) ;
  
#if DEBUG == 2
  if (MPIDB >= 6) 
  {
    cerr << "MPAGMX: Iter. " << l1 
      << " Tol. achieved ";
    cerr.width(9); cerr.precision(1);
    cerr << sk0->r->exponent << endl;
    cerr.width(0); cerr.precision(0);
  }
#endif
  delete sk0; delete sk1;
}

void mp_complex::mpcssx (const mp& a, const mp& pi, mp& x, mp& y, int mpnw)
{
  int k;
  const double cpi = 3.141592653589793, cl2 = 1.4426950408889633;
  
  double t2;
  const int nit = 1;
  mp f1(mpsize(6)); 
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    if (MPIER == 99)  mpabrt();
    mp::zero(x); mp::zero(y);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 5)  mp::mpdeb ("MPCSSX I ", a, mpnw);
#endif
  
  int na = MIN (a.nw, mpnw);
  int ncr = int(pow(2, MPMCR)); // *cast*
  
  //  Check if precision level is too low to justify advanced routine.
  
  if (mpnw <= ncr) 
  {
    mp::mpcssn (a, pi, x, y, mpnw);
#if DEBUG == 2
    if (MPIDB >= 5) 
    {
      mp::mpdeb ("MPCSSX O ", x, mpnw);
      mp::mpdeb ("MPCSSX O ", y, mpnw);
    }
#endif
    return;
  }
  
  //  Check if input is zero
  if (na == 0) 
  {
    x.sign=1; x.nw=1; x.exponent=0; x[0]=1;
    mp::zero(y);
#if DEBUG == 2
    if (MPIDB >= 5) 
    {
      mp::mpdeb ("MPCSSX O ", x, mpnw);
      mp::mpdeb ("MPCSSX O ", y, mpnw);
    }
#endif
    return;
  }
  
  //  Check if Pi has been precomputed.
  
  mp_dpe t1;
  mp::mpmdc (pi, t1);
  if (t1.n != 0 || fabs (t1.a - cpi) > mprx2) 
  {
    if (MPKER[30] != 0) 
    {
      cerr << "*** MPCSSX: PI must be precomputed.\n";
      MPIER = 30;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  f1.nw=1; f1.sign=1; f1.exponent=0;
  f1[0]=1; f1[1]=0;
  int nws = mpnw;
  
  mp_complex *sk0 = createComplex(mpnw+2), 
    *sk1 = createComplex(mpnw+2), 
    *sk2 = createComplex(mpnw+2),
    *sk3 = createComplex(mpnw+2);
  
  //  Reduce argument to between - Pi and Pi.
  
  mp::mpmuld (pi, mp_dpe(2.0), *sk0->r, mpnw);
  mp::mpdivx (a, *sk0->r, *sk1->r, mpnw);
  mp::mpnint (*sk1->r, *sk2->r, mpnw);
  mp::mpmul (*sk2->r, *sk0->r, *sk1->r, mpnw);
  mp::mpsub (a, *sk1->r, *sk0->r,  mpnw);
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  
  t2 = nws;
  int mq = int(cl2 * log (t2) + 1.0 - mprxx); // *cast*
  
  mp::mpeq (f1, *sk2->r, mpnw);
  
  //  Compute initial approximation to [Cos (A), Sin (A)].
  
  mpnw = ncr;
  mp::mpcssn (*sk0->r, pi, *sk3->r, *sk3->i, mpnw);
  int iq = 0;
  
  //  Perform the Newton-Raphson iteration with a dynamically changing precision
  //  level MPNW.
  for (k = MPMCR + 1; k<=mq; k++)
  {
    mpnw = MIN (2 * mpnw, nws);
    bool cont = true;
    while (cont)
    {
      mpangx (*sk3->r, *sk3->i, pi, *sk1->r, mpnw);
      mp::mpsub (*sk0->r, *sk1->r, *sk2->i, mpnw);
      mpcmlx (*sk3, *sk2, *sk1, mpnw);
      mpceq (*sk1, *sk3, mpnw);
      if (k == mq - nit && iq == 0) 
        iq = 1;
      else
        cont = false;
    }
  }
  
  //  The final (cos, sin) result must be normalized to have magnitude 1.
  
  mp::mpsqx (*sk3->r, *sk0->r, mpnw);
  mp::mpsqx (*sk3->i, *sk0->i, mpnw);
  mp::mpadd (*sk0->r, *sk0->i, *sk1->r, mpnw);
  mp::mpsqrx (*sk1->r, *sk2->r, mpnw);
  mp::mpdivx (*sk3->r, *sk2->r, *sk0->r, mpnw);
  mp::mpdivx (*sk3->i, *sk2->r, *sk0->i, mpnw);
  mp::mpeq(*sk0->r, x, mpnw);
  mp::mpeq(*sk0->i, y, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 5) 
  {
    mp::mpdeb ("MPCSSX O ", x, mpnw);
    mp::mpdeb ("MPCSSX O ", y, mpnw);
  }
#endif
  delete sk0; delete sk1; delete sk2; delete sk3;
  
}
